import errno
import os
from io import TextIOWrapper
from typing import Optional

import glm
import numpy
from SpatialTransform import Euler, Pose, Transform

from .bvh import *
from .hierarchy import *


def parseLine(file: TextIOWrapper, lineNumber: int) -> tuple[int, list[str], tuple[TextIOWrapper, int, int, str]]:
    lineNumber += 1
    line = file.readline()
    tokens = line.strip().split()
    debugInfo = (file, lineNumber, len(line) - len(line.lstrip()) + len(tokens[0]), line)
    return (lineNumber, tokens, debugInfo)


def readAsBvh(path: str, loadKeyFrames: bool = True) -> BvhContainer:
    """Deserialize .bvh file into a simple structure."""
    if not os.path.exists(path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

    bvh = BvhContainer()
    with open(path, "r") as file:
        # check for 'HIERARCHY' start
        line, tokens, debugInfo = parseLine(file, 0)
        if not tokens[0] == 'HIERARCHY' and len(tokens) == 1:
            raise SyntaxError('First line must be only "HIERARCHY"', debugInfo)

        # parse joints recursivly
        line, tokens, debugInfo = parseLine(file, line)
        if tokens[0] != 'ROOT':
            raise SyntaxError('First Joint must be defined as "ROOT"', debugInfo)
        bvh.Root = _parseJoint(file, _deserializeJointName(tokens, debugInfo))

        # check for 'MOTION' start
        line, tokens, debugInfo = parseLine(file, line)
        if not tokens[0] == 'MOTION' or not len(tokens) == 1:
            raise SyntaxError('After end of hierarchy must follow "MOTION"', debugInfo)

        # check for frame count
        line, tokens, debugInfo = parseLine(file, line)
        if not tokens[0] == 'Frames:' or not len(tokens) == 2:
            raise SyntaxError('First line of "MOTION" section has to be "Frames: X"', debugInfo)
        else:
            bvh.FrameCount = _deserializeFrameCount(tokens[1:], debugInfo)

        # check for frame rate
        line, tokens, debugInfo = parseLine(file, line)
        if not tokens[0] == 'Frame' or not len(tokens) == 3:
            raise SyntaxError('After frame count must follow "Frame Time"', debugInfo)
        else:
            bvh.FrameTime = _deserializeFrameTime(tokens[2:], debugInfo)

        # parse motion data
        if loadKeyFrames:
            for _ in range(bvh.FrameCount):
                line, tokens, debugInfo = parseLine(file, line)
                _deserializeMotion(bvh.Root, _deserializeKeyframe(tokens, debugInfo))

        return bvh


def convertBvhToHierarchy(bvh: BvhJoint) -> Joint:
    """Converts a deserialized bvh structure into a joint hierarchy."""
    # copy data into a joint
    restPose = Transform(name='RestPose', position=bvh.Offset, rotation=bvh.getRotation())
    joint = Joint(bvh.Name, restPose=restPose)
    for frame, key in enumerate(bvh.Keyframes):
        joint.setKeyframe(frame, Transform.fromPose(key), keep=None)

    # correct bvh keyframe data
    for frame, key in joint.Keyframes:
        # project offset into rest pose because it is given as overwrite and does not include the rest pose rotation.
        key.Position = joint.RestPose.SpaceInverse * key.Position

        # the rotation is already given as difference, but the multiplication order is switched.
        key.Rotation = key.Rotation * joint.RestPose.Rotation
        key.Rotation = (glm.inverse(joint.RestPose.Rotation) * key.Rotation)

    for child in bvh.Children:
        # correct the rest pose, because its given without the parents rest pose rotation.
        childJoint = convertBvhToHierarchy(child)
        childJoint.RestPose.Position = glm.inverse(joint.RestPose.Rotation) * childJoint.RestPose.Position
        childJoint.RestPose.Rotation = glm.inverse(joint.RestPose.Rotation) * childJoint.RestPose.Rotation
        joint.attach(childJoint, keep=None)

    return joint


def convertHierarchyToBvh(joint: Joint, frames: int, worldSpace: Optional[Pose] = None) -> BvhJoint:
    """Converts a joint structure into a deseralized bvh structure."""

    if worldSpace is None:
        worldSpace = Pose()

    bvh = BvhJoint(joint.Name)
    bvh.Offset = worldSpace.Space * joint.RestPose.Position
    bvh.EndSite = (worldSpace.Space * (0, 1, 0)) * glm.length(joint.RestPose.Position) * 0.3
    bvh.Keyframes = [joint.getKeyframe(frame).toPose(worldSpace=False) for frame in range(frames)]

    worldSpace.Rotation = worldSpace.Rotation * joint.RestPose.Rotation
    worldSpace.Scale = worldSpace.Scale * joint.RestPose.Scale

    if 1e-02 < glm.l1Norm(sum([glm.abs(pose.Position) for pose in bvh.Keyframes])):
        bvh.Channels.extend(['Xposition', 'Yposition', 'Zposition'])

    if 1e-02 < (sum([sum([abs(d) for d in (pose.Rotation - glm.quat()).to_list()]) for pose in bvh.Keyframes])):
        bvh.Channels.extend(['Zrotation', 'Xrotation', 'Yrotation'])

    # convert data to bvh
    for key in bvh.Keyframes:
        key.Position = bvh.Offset + (worldSpace.Space * key.Position)
        key.Rotation = worldSpace.Rotation * key.Rotation
        key.Rotation = (key.Rotation * glm.inverse(worldSpace.Rotation))

    # add children
    for child in joint.Children:
        childBvh = convertHierarchyToBvh(child, frames, worldSpace.duplicate())
        bvh.Children.append(childBvh)

    return bvh


def readAsHierarchy(path: str, loadKeyFrames: bool = True) -> Joint:
    """Deserialize a .bvh file into a joint hierarchy."""
    return convertBvhToHierarchy(readAsBvh(path, loadKeyFrames).Root).loadRestPose(recursive=True)


def _parseJoint(file: TextIOWrapper, name: str, line: int = 0) -> BvhJoint:
    # check for open bracket
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != '{' or len(tokens) > 1:
        raise SyntaxError('Joint definition must start with an opening bracket', debugInfo)

    # create joint and read channels and offeset
    joint = BvhJoint(name)
    joint.Offset = _deserializeOffset(file, line)
    joint.Channels = _deserializeChannles(file, line)

    # check for definition end or child joints or end site info
    while (True):
        line, tokens, debugInfo = parseLine(file, line)
        if tokens[0] == 'JOINT':
            joint.Children.append(_parseJoint(file, _deserializeJointName(tokens, debugInfo), line))
        elif tokens[0] == 'End':
            joint.EndSite = _deserializeEndSite(file, line)
        elif tokens[0] == '}':
            break
        else:
            raise SyntaxError('Joint definition must end with an child joint, end site or closing bracket', debugInfo)

    # return deserialized joint root pose
    return joint


def _deserializeJointName(tokens: list[str], debugInfo: tuple[TextIOWrapper, int, int, str]) -> str:
    if not isinstance(tokens, list) and len(tokens) != 2:
        raise SyntaxError('Joint header must be 2-dimensional tuple', debugInfo)
    return tokens[1]


def _deserializeOffset(file: TextIOWrapper, line: int) -> glm.vec3:
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != 'OFFSET':
        raise SyntaxError('Expected OFFSET definition for joint', debugInfo)
    if not isinstance(tokens, list) or len(tokens[1:]) != 3:
        raise SyntaxError('Offset must be a 3-part tuple', debugInfo)
    try:
        return glm.vec3(list(map(float, tokens[1:])))
    except ValueError:
        raise SyntaxError('Offset must be numerics only', debugInfo)


def _deserializeChannles(file: TextIOWrapper, line: int) -> list[str]:
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != 'CHANNELS':
        raise SyntaxError('Expected CHANNELS definition for joint', debugInfo)
    if not isinstance(tokens, list) or len(tokens) < 2:
        raise SyntaxError('Channels must be at least a 2-part tuple', debugInfo)
    try:
        channelCount = int(tokens[1])
    except ValueError:
        raise SyntaxError('Channel count must be numerical', debugInfo)
    if channelCount != len(tokens[2:]):
        raise SyntaxError('Channel count mismatch with labels', debugInfo)
    return tokens[2:]


def _deserializeEndSite(file: TextIOWrapper, line: int) -> glm.vec3:
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != '{':
        raise SyntaxError('End Site definition must start with an opening bracket', debugInfo)
    result = _deserializeOffset(file, line)
    line, tokens, debugInfo = parseLine(file, line)
    if tokens[0] != '}':
        raise SyntaxError('End Site definition must start with an opening bracket', debugInfo)
    return result


def _deserializeFrameTime(data: list, debugInfo: tuple) -> float:
    if not isinstance(data, list) or len(data) != 1:
        raise SyntaxError('Frame time must be a 1-dimensional tuple', debugInfo)
    try:
        return float(data[0])
    except ValueError:
        raise SyntaxError('Frame time be numerical', debugInfo)


def _deserializeFrameCount(data: list, debugInfo: tuple) -> float:
    if not isinstance(data, list) or len(data) != 1:
        raise SyntaxError('Frame count must be a 1-dimensional tuple', debugInfo)
    try:
        return int(data[0])
    except ValueError:
        raise SyntaxError('Frame time be numerical', debugInfo)


def _deserializeKeyframe(data: list, debugInfo: tuple) -> numpy.ndarray:
    try:
        return numpy.array(list(map(float, data)))
    except ValueError:
        raise SyntaxError('Keyframe must be numerics only', debugInfo)


def _deserializeMotion(joint: BvhJoint, data: numpy.ndarray, index=0) -> int:
    position = glm.vec3(joint.Offset)
    rotation = glm.vec3(0)
    rotOrder = ''

    for channel in joint.Channels:
        if 'Xposition' == channel: position.x = data[index]
        elif 'Yposition' == channel: position.y = data[index]
        elif 'Zposition' == channel: position.z = data[index]
        elif 'Xrotation' == channel: rotation.x = data[index]; rotOrder += 'X'
        elif 'Yrotation' == channel: rotation.y = data[index]; rotOrder += 'Y'
        elif 'Zrotation' == channel: rotation.z = data[index]; rotOrder += 'Z'
        else: index -= 1
        index += 1

    rotation = Euler.toQuatFrom(glm.radians(rotation), order=rotOrder, extrinsic=False)
    joint.Keyframes.append(Pose(position, rotation))

    for child in joint.Children:
        index = _deserializeMotion(child, data, index)
    return index


def writeBvh(path: str, bvh: BvhContainer, percision: int = 9) -> None:
    with open(path, "w") as file:
        file.write('HIERARCHY\n')
        writeJoint(file, bvh.Root, 0, True, percision)

        file.write('MOTION\n')
        file.write(f'Frames: {bvh.FrameCount}\n')
        file.write(f'Frame Time: {bvh.FrameTime}\n')

        for frame in range(bvh.FrameCount):
            writeMotion(file, bvh.Root, frame, percision)
            file.write('\n')


def writeHierarchy(path: str, root: Joint, frameTime: float, frames: int = None, percision: int = 9) -> None:
    """Creates an .bvh file from the given hierarchy.
    - frameTime defines the FPS
    - IF frames is None -> THe whole animation is written.
    - If frames is set -> The Animation is written from frame 0 to frame.
    - percision limits the percision of floating numbers be written.
    - Data will be overwritten if the file already exists"""
    frames = (root.getKeyframeRange()[1] + 1) if frames is None else frames
    container = BvhContainer(convertHierarchyToBvh(root, frames + 1), frames, frameTime)
    writeBvh(path, container, percision)


def writeJoint(file: TextIOWrapper, joint: BvhJoint, indent: int, isFirst: bool, percision: int) -> None:
    file.write(f'{"  "*indent}{"ROOT" if isFirst else "JOINT"} {joint.Name}\n')
    file.write(f'{"  "*indent}{{\n')
    file.write(f'{"  "*(indent+1)}OFFSET ')
    file.write(f'{round(joint.Offset.x, percision)} ')
    file.write(f'{round(joint.Offset.y, percision)} ')
    file.write(f'{round(joint.Offset.z, percision)}\n')
    file.write(f'{"  "*(indent+1)}CHANNELS {len(joint.Channels)} {" ".join(joint.Channels)}\n')
    if len(joint.Children) > 0:
        for child in joint.Children:
            writeJoint(file, child, indent + 1, False, percision)
    else:
        file.write(f'{"  "*(indent+1)}End Site\n')
        file.write(f'{"  "*(indent+1)}{{\n')
        file.write(f'{"  "*(indent+2)}OFFSET ')
        file.write(f'{round(joint.EndSite.x, percision)} ')
        file.write(f'{round(joint.EndSite.y, percision)} ')
        file.write(f'{round(joint.EndSite.z, percision)}\n')
        file.write(f'{"  "*(indent+1)}}}\n')
    file.write(f'{"  "*indent}}}\n')


def writeMotion(file: TextIOWrapper, joint: BvhJoint, frame: int, percision: int) -> None:
    rotOrder = ''.join([rot[0] for rot in joint.Channels if rot[1:] == 'rotation'])
    if 'Z' not in rotOrder: rotOrder += 'Z'
    if 'X' not in rotOrder: rotOrder += 'X'
    if 'Y' not in rotOrder: rotOrder += 'Y'

    rotation = joint.Keyframes[frame].getEuler(rotOrder, extrinsic=False)
    position = joint.Keyframes[frame].Position

    for channel in joint.Channels:
        if 'Xposition' == channel: file.write(f'{round(position.x, percision)} '); continue
        if 'Yposition' == channel: file.write(f'{round(position.y, percision)} '); continue
        if 'Zposition' == channel: file.write(f'{round(position.z, percision)} '); continue
        if 'Xrotation' == channel: file.write(f'{round(rotation.x, percision)} '); continue
        if 'Yrotation' == channel: file.write(f'{round(rotation.y, percision)} '); continue
        if 'Zrotation' == channel: file.write(f'{round(rotation.z, percision)} '); continue

    for child in joint.Children:
        writeMotion(file, child, frame, percision)



import unittest
import bvhio
import glm
from utils import *

class Reading(unittest.TestCase):
    def setUp(self):
        self.instance = bvhio.readAsHierarchy('bvhio/tests/example.bvh')
        self.joints = [ # for j, i, d in root.layout(): print(f"( {d}, {len(j.Children)}, {len(j.Keyframes)}, {j.KeyframeRange}, '{j.Name}', ),")
            ( 0, 3, 2, (0, 1), 'Hips', ),
            ( 1, 3, 2, (0, 1), 'Chest', ),
            ( 2, 1, 2, (0, 1), 'Neck', ),
            ( 3, 0, 2, (0, 1), 'Head', ),
            ( 2, 1, 2, (0, 1), 'LeftCollar', ),
            ( 3, 1, 2, (0, 1), 'LeftUpArm', ),
            ( 4, 1, 2, (0, 1), 'LeftLowArm', ),
            ( 5, 0, 2, (0, 1), 'LeftHand', ),
            ( 2, 1, 2, (0, 1), 'RightCollar', ),
            ( 3, 1, 2, (0, 1), 'RightUpArm', ),
            ( 4, 1, 2, (0, 1), 'RightLowArm', ),
            ( 5, 0, 2, (0, 1), 'RightHand', ),
            ( 1, 1, 2, (0, 1), 'LeftUpLeg', ),
            ( 2, 1, 2, (0, 1), 'LeftLowLeg', ),
            ( 3, 0, 2, (0, 1), 'LeftFoot', ),
            ( 1, 1, 2, (0, 1), 'RightUpLeg', ),
            ( 2, 1, 2, (0, 1), 'RightLowLeg', ),
            ( 3, 0, 2, (0, 1), 'RightFoot', ),
        ]
        self.restPoseLocal = [ # for j, i, d in root.layout(): print(f"( glm.{j.Position}, glm.{j.Rotation}, glm.{j.Scale}, ),")
            ( glm.vec3(            0,            0,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         5.21,            0 ), glm.quat(     0.999333,    0.0365139,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,      18.6003,     -1.36106 ), glm.quat(     0.999333,   -0.0365139,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         5.45,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         1.12,      16.3232,     0.680562 ), glm.quat(     0.706635,   -0.0258192,   -0.0258192,    -0.706635 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3( -4.76837e-07,         5.54,            0 ), glm.quat(    -0.707107,            0,            0,     0.707107 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        11.96,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         9.93,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -1.12,      16.3232,     0.680562 ), glm.quat(     0.706635,   -0.0258192,    0.0258192,     0.706635 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(  9.53674e-07,         6.07,            0 ), glm.quat(     0.707107,            0,            0,     0.707107 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        11.82,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        10.65,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         3.91,            0,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        18.34,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        17.37,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -3.91,            0,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        17.63,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        17.14,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
        ]
        self.restPoseWorld = [ # for j, i, d in root.layout(): print(f"( glm.{j.PositionWorld}, glm.{j.RotationWorld}, glm.{j.ScaleWorld}, ),")
            ( glm.vec3(            0,            0,            0 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         5.21,            0 ), glm.quat(     0.999333,    0.0365139,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        23.86,  1.19209e-07 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        29.31,  1.19209e-07 ), glm.quat(            1,            0,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         1.12,        21.44,         1.87 ), glm.quat(     0.707107,            0,            0,    -0.707107 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         6.66,        21.44,         1.87 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         6.66,         9.48,         1.87 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         6.66,    -0.450002,         1.87 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -1.12,        21.44,         1.87 ), glm.quat(     0.707107,            0,            0,     0.707107 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -7.19,        21.44,         1.87 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -7.19,         9.62,         1.87 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -7.19,        -1.03,         1.87 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         3.91,            0,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         3.91,       -18.34,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         3.91,       -35.71,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -3.91,            0,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -3.91,       -17.63,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -3.91,       -34.77,            0 ), glm.quat(            0,            0,            0,            1 ), glm.vec3(            1,            1,            1 ), ),
        ]
        self.pose0Local = [
            ( glm.vec3(         8.03,        35.01,        88.36 ), glm.quat(     0.131166,   -0.0117277,    -0.982546,    -0.131386 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         5.21,            0 ), glm.quat(     0.905976,     0.390515,    -0.159062,    0.0374775 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,      18.6003,     -1.36106 ), glm.quat(     0.938613,     0.338059,    0.0279395,     0.062775 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         5.45,            0 ), glm.quat(     0.933054,    -0.351756,    0.0585907,   -0.0473785 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         1.12,      16.3232,     0.680562 ), glm.quat(     0.760453,   -0.0960347,    0.0337522,    -0.641365 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3( -4.76837e-07,         5.54,            0 ), glm.quat(    -0.976203,    -0.202105,   -0.0323696,   -0.0716515 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        11.96,            0 ), glm.quat(     0.694325,     0.714836,    0.0378194,   -0.0741041 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,         9.93,            0 ), glm.quat(     0.999982, -0.000261795, -1.57639e-06,   0.00602135 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -1.12,      16.3232,     0.680562 ), glm.quat(     0.781762,    -0.100736,   -0.0338954,     0.614452 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(  9.53674e-07,         6.07,            0 ), glm.quat(     0.614408,    0.0440009,     0.778943,     0.117536 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        11.82,            0 ), glm.quat(     0.762976,     0.606189,    -0.177084,     0.138002 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        10.65,            0 ), glm.quat(     0.999888,  -0.00602318, -9.20184e-05,   -0.0136991 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(         3.91,            0,            0 ), glm.quat(    -0.117674,    -0.024446,    -0.202143,     0.971953 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        18.34,            0 ), glm.quat(     0.894013,    -0.411485,    -0.106536,     0.141674 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        17.37,            0 ), glm.quat(     0.999951,   0.00994821,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(        -3.91,            0,            0 ), glm.quat(     0.141065,   -0.0135394,    0.0944943,     0.985387 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        17.63,            0 ), glm.quat(     0.883393,      -0.4543,     0.109673,    0.0346498 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(            0,        17.14,            0 ), glm.quat(     0.978238,     0.207485,            0,            0 ), glm.vec3(            1,            1,            1 ), ),
        ]
        self.pose0World = [
            ( glm.vec3(         8.03,        35.01,        88.36 ), glm.quat(     0.131166,   -0.0117277,    -0.982546,    -0.131386 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      8.32964,      40.0387,      89.6891 ), glm.quat(    -0.027949,   -0.0171246,    -0.961895,     0.271448 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      9.16411,       56.599,      81.1521 ), glm.quat(   -0.0106095,   -0.0934888,    -0.810788,     0.577729 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      10.0571,      58.3157,      76.0571 ), glm.quat(    0.0320922,   -0.0789339,    -0.964779,     0.248878 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      8.02774,      53.6106,      80.5308 ), glm.quat(     0.183665,     0.597426,    -0.769471,     0.131396 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      2.66686,      55.0047,      80.6263 ), glm.quat(   -0.0740443,    -0.560941,     0.761465,    -0.316281 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(     -8.11043,      57.0454,       75.859 ), glm.quat(     0.297335,    -0.486871,     0.258247,    -0.779652 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(     -6.00359,      50.1956,      68.9853 ), glm.quat(     0.301897,    -0.485386,     0.261377,    -0.777779 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      10.2629,      53.5708,      80.6721 ), glm.quat(     -0.22297,     -0.59241,    -0.767848,     0.098717 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      16.0524,       55.262,      81.3554 ), glm.quat(     0.475579,    -0.540937,     -0.57148,    -0.393222 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      27.7812,      56.5094,      80.5862 ), glm.quat(     0.643831,     -0.27293,    -0.683958,     0.207828 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      28.9073,      64.6527,      73.8156 ), glm.quat(     0.644899,    -0.267389,    -0.688931,      0.19489 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      4.25561,      34.9653,      89.3799 ), glm.quat(   -0.0866354,    -0.983374,     0.103717,     0.121299 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(     0.900007,      17.2952,      92.9663 ), glm.quat(    -0.488232,    -0.815883,      0.19136,     0.243612 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(    -0.391913,      9.47834,      108.424 ), glm.quat(    -0.480091,    -0.820699,     0.193774,     0.241696 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      11.8044,      35.0547,      87.3401 ), glm.quat(     0.240655,    -0.959203,    -0.112873,    0.0963038 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      14.8047,       19.916,      78.8176 ), glm.quat(    -0.214131,    -0.971155,   -0.0838329,   -0.0630639 ), glm.vec3(            1,            1,            1 ), ),
            ( glm.vec3(      17.1327,      4.58869,      86.1275 ), glm.quat(  -0.00797081,     -0.99445,   -0.0950934,   -0.0442974 ), glm.vec3(            1,            1,            1 ), ),
        ]

    def test_JointCount(self):
        self.assertEqual(len(self.joints), len(self.instance.layout()))

    def test_Name(self):
        for j, i, d in self.instance.layout():
            self.assertEqual(j.Name, self.joints[i][-1])

    def test_Depth(self):
        for j, i, d in self.instance.layout():
            self.assertEqual(d, self.joints[i][0])

    def test_Children(self):
        for j, i, d in self.instance.layout():
            self.assertEqual(len(j.Children), self.joints[i][1])

    def test_Keyframes(self):
        for j, i, d in self.instance.layout():
            self.assertEqual(2, self.joints[i][2])

    def test_KeyframeRange(self):
        for j, i, d in self.instance.layout():
            self.assertEqual(j.getKeyframeRange(includeChildren=True), self.joints[i][3])

    def test_PositionsRestPose(self):
        for j, i, d in self.instance.loadRestPose(recursive=True).layout():
            self.assertGreater(1e-04, deviationPosition(j.Position, self.restPoseLocal[i][0]))
            self.assertGreater(1e-04, deviationPosition(j.PositionWorld, self.restPoseWorld[i][0]))

    def test_RotationsRestPose(self):
        for j, i, d in self.instance.loadRestPose(recursive=True).layout():
            self.assertGreater(1e-06, deviationQuaternion(j.Rotation, self.restPoseLocal[i][1]))
            self.assertGreater(1e-06, deviationQuaternion(j.RotationWorld, self.restPoseWorld[i][1]))

    def test_ScalesRestPose(self):
        for j, i, d in self.instance.loadRestPose(recursive=True).layout():
            self.assertGreater(1e-06,  glm.length(j.Scale - self.restPoseLocal[i][2]))
            self.assertGreater(1e-06,  glm.length(j.ScaleWorld - self.restPoseWorld[i][2]))

    def test_PositionsPose0(self):
        for j, i, d in self.instance.loadPose(0, recursive=True).layout():
            self.assertGreater(1e-04, deviationPosition(j.Position, self.pose0Local[i][0]))
            self.assertGreater(1e-03, deviationPosition(j.PositionWorld, self.pose0World[i][0]))

    def test_RotationsPose0(self):
        for j, i, d in self.instance.loadPose(0, recursive=True).layout():
            self.assertGreater(1e-05, deviationQuaternion(j.Rotation, self.pose0Local[i][1]))
            self.assertGreater(1e-05, deviationQuaternion(j.RotationWorld, self.pose0World[i][1]))

    def test_ScalesPose0(self):
        for j, i, d in self.instance.loadPose(0, recursive=True).layout():
            self.assertGreater(1e-06,  glm.length(j.Scale - self.pose0Local[i][2]))
            self.assertGreater(1e-06,  glm.length(j.ScaleWorld - self.pose0World[i][2]))

class Methods(unittest.TestCase):
    def test_write_read_RestPose(self):
        instance = bvhio.readAsHierarchy('bvhio/tests/example.bvh')
        instance.loadRestPose(recursive=True)
        restPose = [(joint.PositionWorld, joint.RotationWorld, joint.ScaleWorld) for joint, _ , _ in instance.layout()]
        instance.loadPose(0, recursive=True)
        animPose = [(joint.PositionWorld, joint.RotationWorld, joint.ScaleWorld) for joint, _ , _ in instance.layout()]

        instance.loadRestPose(recursive=True)
        instance.writeRestPose(recursive=True, keep=None)
        instance.loadRestPose(recursive=True)
        for j, i, d in instance.layout():
            self.assertGreater(1e-04, deviationPosition(restPose[i][0], j.PositionWorld))
            self.assertGreater(1e-04, deviationQuaternion(restPose[i][1], j.RotationWorld))
            self.assertGreater(1e-04, deviationScale(restPose[i][2], j.ScaleWorld))
        instance.loadPose(0, recursive=True)
        for j, i, d in instance.layout():
            self.assertGreater(1e-04, deviationPosition(animPose[i][0], j.PositionWorld))
            self.assertGreater(1e-04, deviationQuaternion(animPose[i][1], j.RotationWorld))
            self.assertGreater(1e-04, deviationScale(animPose[i][2], j.ScaleWorld))

        instance.loadRestPose(recursive=True)
        instance.writeRestPose(recursive=True, keep=['position', 'rotation', 'scale'])
        instance.loadRestPose(recursive=True)
        for j, i, d in instance.layout():
            self.assertGreater(1e-04, deviationPosition(restPose[i][0], j.PositionWorld))
            self.assertGreater(1e-04, deviationQuaternion(restPose[i][1], j.RotationWorld))
            self.assertGreater(1e-04, deviationScale(restPose[i][2], j.ScaleWorld))
        instance.loadPose(0, recursive=True)
        for j, i, d in instance.layout():
            self.assertGreater(1e-04, deviationPosition(animPose[i][0], j.PositionWorld))
            self.assertGreater(1e-04, deviationQuaternion(animPose[i][1], j.RotationWorld))
            self.assertGreater(1e-04, deviationScale(animPose[i][2], j.ScaleWorld))

    def test_write_read_Pose(self):
        instance = bvhio.readAsHierarchy('bvhio/tests/example.bvh')
        instance.loadRestPose(recursive=True)
        restPose = [(joint.PositionWorld, joint.RotationWorld, joint.ScaleWorld) for joint, _ , _ in instance.layout()]

        for frame in range(len(instance.Keyframes)):
            instance.loadPose(frame, recursive=True)
            pose = [(joint.PositionWorld, joint.RotationWorld, joint.ScaleWorld) for joint, _ , _ in instance.layout()]

            instance.writePose(frame, recursive=True)
            instance.loadPose(frame, recursive=True)
            for j, i, d in instance.layout():
                self.assertGreater(1e-04, deviationPosition(pose[i][0], j.PositionWorld))
                self.assertGreater(1e-04, deviationQuaternion(pose[i][1], j.RotationWorld))
                self.assertGreater(1e-04, deviationScale(pose[i][2], j.ScaleWorld))
            instance.loadRestPose(recursive=True)
            for j, i, d in instance.layout():
                self.assertGreater(1e-04, deviationPosition(restPose[i][0], j.PositionWorld))
                self.assertGreater(1e-04, deviationQuaternion(restPose[i][1], j.RotationWorld))
                self.assertGreater(1e-04, deviationScale(restPose[i][2], j.ScaleWorld))

    def test_roll(self):
        instance = bvhio.readAsHierarchy('bvhio/tests/example.bvh')
        instance.loadRestPose(recursive=True)
        childPose = [(joint.PositionWorld, joint.RotationWorld, joint.ScaleWorld) for joint in instance.Children]

        for step in range(72):
            instance.roll(step*5, recursive=False)
            for i, child in enumerate(instance.Children):
                self.assertGreater(1e-04, deviationPosition(childPose[i][0], child.PositionWorld))
                self.assertGreater(1e-04, deviationQuaternion(childPose[i][1], child.RotationWorld))
                self.assertGreater(1e-04, deviationScale(childPose[i][2], child.ScaleWorld))

        self.assertTrue(True)



import glm
from SpatialTransform import Pose


class BvhJoint:
    """Data structure for the bvh skeleton definition. Contains the attributes as in the BVH file.

    Keyframes contain the motion data."""
    Name: str
    Offset: glm.vec3
    EndSite: glm.vec3
    Keyframes: list[Pose]
    Channels: list[str]
    Children: list["BvhJoint"]

    def __init__(self, name: str, offset: glm.vec3 = None) -> None:
        self.Name = name
        self.Offset = glm.vec3() if offset is None else glm.vec3(offset)
        self.EndSite = glm.vec3(0, 1, 0)
        self.Keyframes = []
        self.Channels = []
        self.Children = []

    def __repr__(self) -> str:
        return (f"{self.Name}")

    def __str__(self) -> str:
        return self.__repr__()

    def getTip(self) -> glm.vec3:
        """Calculates the tip of the defined bone.

        The tip position is the average position of all children in local space without rotation.
        If there is no child joint, the EndSite value will be returned."""
        children = len(self.Children)

        if children == 1:
            return self.Children[0].Offset
        elif children > 1:
            return sum(child.Offset for child in self.Children) / children
        else:
            return self.EndSite if glm.length(self.EndSite) > 0.001 else glm.vec3(0, 1, 0)

    def getLength(self) -> float:
        """Length of the bone which is based on the tip."""
        return glm.length(self.getTip())

    def getRotation(self) -> glm.quat:
        """calculate the bone rotation based on the tip."""
        dir = glm.normalize(self.getTip())
        axs = glm.vec3(0, 1, 0)
        dot = glm.dot(axs, dir)

        if dot < -0.9999: return glm.quat(0, 0, 0, 1)
        if dot > +0.9999: return glm.quat(1, 0, 0, 0)
        return glm.angleAxis(glm.acos(dot), glm.normalize(glm.cross(axs, dir)))

    def layout(self, index: int = 0, depth: int = 0) -> list[tuple["BvhJoint", int, int]]:
        """Returns the hierarchical layout of this joint and its children recursivly."""
        result = [[self, index, depth]]
        for child in self.Children:
            result.extend(child.layout(result[-1][1] + 1, depth + 1))
        return result



import glm
import bisect
from SpatialTransform import Transform, Pose


class Joint(Transform):
    """Spatial definition of an linear space with position, rotation and scale.
    - Bone alignment is expected to be along the Y+ axis.
    - Space is defined as right handed where Y+:up and X+:right and Z-:forward.
    - Positive rotations are counter clockwise.
    - The animation is a cualculation of ``Pose = RestPose + Keyframe``
    - The RestPose and Keyframe data is in local space only.
    - The method ``readPose()`` combines the RestPose and Keframes."""

    @property
    def Parent(self) -> "Joint":
        return self._Parent

    @property
    def Children(self) -> list["Joint"]:
        return self._Children

    @property
    def CurrentFrame(self) -> int:
        """Latest frame index that has been read with readPose()."""
        return self._CurrentFrame

    @property
    def Keyframes(self) -> list[tuple[int, Transform]]:
        """Animation data for the joint. A keyframe holds the change of local properties in relation to the rest pose, so that ``Pose = RestPose + Keyframe``.
    - The first element in the tuple is the frame id and the second element are the local keyframe properties.
    - This is an ordered list by the frame id.
    - Negative frame ids should not exist."""
        return self._Keyframes

    @Keyframes.setter
    def Keyframes(self, value: list[tuple[int, Transform]]) -> None:
        self._Keyframes = list(value)
        self._KeyframeMap = {frame: key for frame, key in self._Keyframes}
        self.RestPose.clearChildren(keep=[None])
        self.RestPose.attach(*[key for frame, key in value], keep=[None])

    @property
    def RestPose(self) -> Transform:
        """Pose without any keyframe applied. The common T-Pose would go here. This Pose is the base for all animation data."""
        return self._RestPose

    @RestPose.setter
    def RestPose(self, value: Pose) -> None:
        self._RestPose = value.duplicate()

    def __init__(
            self, name: str = None,
            position: glm.vec3 = None,
            rotation: glm.quat = None,
            scale: glm.vec3 = None,
            restPose: Transform = None,
            keyFrames: list[tuple[int, Transform]] = None) -> None:

        super().__init__(name, position, rotation, scale)
        self._Parent: "Joint" = None
        self._Children: list["Joint"] = []

        self._RestPose: Transform = Transform(name='RestPose') if restPose is None else restPose
        self._Keyframes: list[tuple[int, Transform]] = [] if keyFrames is None else keyFrames
        self._KeyframeMap = {frame: key for frame, key in self._Keyframes}
        self._CurrentFrame = -1

    def __findFrameIndex(self, frame: int):
        if not self.Keyframes:
            return 0

        if frame > self.Keyframes[-1][0]:
            return len(self.Keyframes)
               
        return bisect.bisect_left([key[0] for key in self.Keyframes], frame)

    def __insertKeyframe(self, frame: int, key: Transform) -> None:
        bisect.insort(self.Keyframes, (frame, key))
        self._KeyframeMap[frame] = key

    def getKeyframe(self, frame: int) -> Transform:
        """Returns the pose at the given frame id.
        - If the frame number is negative, it will look for the n-th frame from the end.
        - If there are no keyframes, the joint propetries will not change.
        - If the frame id is out of the keyframe length, the nearest keyframe propetires are used.
        - If the frame id is between two keyframes, pose properties are linearly interpolated."""
        if len(self.Keyframes) == 0:
            key = Transform(name=f'Key {frame} (placeholder)')
            self.RestPose.duplicate(recursive=False).attach(key, keep=None)
            return key

        if frame < 0: frame = max(0, self.getKeyframeRange(includeChildren=False)[1] + 1 - frame)

        existingFrame = self._KeyframeMap.get(frame)

        if existingFrame:
            return existingFrame
        
        # pose definition
        index = self.__findFrameIndex(frame)
        if index == len(self.Keyframes):
            # index is bigger than last frame, take last key
            return self.Keyframes[-1][1]
        elif self.Keyframes[index][0] == frame:
            # index matches a keyframe
            return self.Keyframes[index][1]
        else:
            if index == 0:
                # index is smaller than first frame, take first key
                return self.Keyframes[-1][1]
            else:
                # index is in between two keyframes, interpolate
                before = self.Keyframes[index - 1]
                after = self.Keyframes[index]
                weight = (before[0] + frame) / after[0]

                restPoseCopy = self.RestPose.duplicate(recursive=False).attach(
                    Transform(
                        name=f'Key {frame} (interpolated)',
                        position=glm.lerp(before[1].Position, after[1].Position, weight),
                        rotation=glm.lerp(before[1].Rotation, after[1].Rotation, weight),
                        scale=glm.lerp(before[1].Scale, after[1].Scale, weight)
                    ), keep=None
                )
                return restPoseCopy.Children[0]

    def setKeyframe(self, frame: int, pose: Transform, keep: list[str] = ['position', 'rotation', 'scale']) -> "Joint":
        """Inserts the given pose to the the keyframes.
        - If there is already a keyframe at the frame id, it will be overwritten.
        - If the frame number is negative, it counts as the n-th frame from the end.
        - This pose is added later to the rest pose to calculate the final animation."""
        if frame < 0: frame = max(0, self.getKeyframeRange(includeChildren=False)[1] + 1 - frame)
        index = self.__findFrameIndex(frame)

        if index == len(self.Keyframes) or self.Keyframes[index][0] != frame:
            newKey = Transform(name=f'Key {frame}', position=pose.Position, rotation=pose.Rotation, scale=pose.Scale)
            self.__insertKeyframe(frame, newKey)
            self.RestPose.attach(newKey, keep=keep)
        else:
            if 'position' in keep: self.Keyframes[index][1].PositionWorld = pose.Position
            else: self.Keyframes[index][1].Position = pose.Position

            if 'rotation' in keep: self.Keyframes[index][1].RotationWorld = pose.Rotation
            else: self.Keyframes[index][1].Rotation = pose.Rotation

            if 'scale' in keep: self.Keyframes[index][1].ScaleWorld = pose.Scale
            else: self.Keyframes[index][1].Scale = pose.Scale

        return self

    def removeKeyframe(self, frame: int, recursive: bool = False) -> "Joint":
        """Removes the keyframe, if it exists, from the keyframe list.
        - If recursive is True -> Child joints do also load their rest pose.

        Returns itself."""
        index = self.__findFrameIndex(frame)

        if index != len(self.Keyframes) and self.Keyframes[index][0] == frame:
            self.Keyframes.pop(index)[1].clearParent(keep=None)

        if recursive:
            for child in self.Children:
                child.removeKeyframe(frame=frame, recursive=True)

        return self

    def loadKeyframe(self, frame: int, recursive: bool = True, use: bool = ['position', 'rotation', 'scale']) -> "Joint":
        """Loads the pose at the given frame into the transforms properties.
        - This is the animation data without rest pose.

        Returns itself."""
        key = self.getKeyframe(frame)

        if 'position' in use: self.Position = key.Position
        if 'rotation' in use: self.Rotation = key.Rotation
        if 'scale' in use: self.Scale = key.Scale

        if recursive:
            for child in self.Children:
                child.loadKeyframe(frame=frame, recursive=True, use=use)

        return self

    def loadRestPose(self, recursive: bool = True, use: bool = ['position', 'rotation', 'scale']) -> "Joint":
        """Sets joint properties to the rest pose.
        - If recursive is True -> Child joints do also load their rest pose.

        Returns itself."""
        # self alignment
        if 'position' in use: self.Position = self.RestPose.Position
        if 'rotation' in use: self.Rotation = self.RestPose.Rotation
        if 'scale' in use: self.Scale = self.RestPose.Scale

        # recursion
        if recursive:
            for child in self.Children:
                child.loadRestPose(recursive=True, use=use)

        return self

    def writeRestPose(self, recursive: bool = True, keep: list[str] = ['position', 'rotation', 'scale']) -> "Joint":
        """Sets the rest pose to the current joint properties.
        - If keep contains properties -> The Keyframes are modified to keep its spatial algiment in world space.
        - If keep is None or empty -> Keyframes do not change and thus the animation will change.
        - If recursive is True -> Child joints do also load their rest pose.

        Returns itself."""
        # remove change in rest pose from keyframes
        if keep:
            for frame, key in self.Keyframes:
                if 'position' in keep: key.Position = self.SpaceInverse * key.PositionWorld
                if 'rotation' in keep: key.Rotation = glm.inverse(self.Rotation) * key.RotationWorld
                if 'scale' in keep: key.Scale = (key.Scale / self.ScaleWorld)

        # write rest pose
        self.RestPose.Position = self.Position
        self.RestPose.Rotation = self.Rotation
        self.RestPose.Scale = self.Scale

        # recursion
        if recursive:
            for child in self.Children:
                child.writeRestPose(recursive=True, keep=keep)

        return self

    def loadPose(self, frame: int, recursive: bool = True, use: bool = ['position', 'rotation', 'scale']) -> "Joint":
        """Sets joint properties to the animation at the given frame id. The animation is defined as 'Pose = RestPose + Keyframe'.
        - If the frame number is negative, it will look for the n-th frame from the end.
        - If there are no keyframes, the joint propetries will not change.
        - If the frame id is out of the keyframe length, the nearest keyframe propetires are used.
        - If the frame id is between two keyframes, pose properties are linearly interpolated.
        - If recursive is True -> Child joints do also load their pose.

        Returns itself."""
        # get animation data
        key = self.getKeyframe(frame)

        # set animation pose, world space includes the transform from the rest pose
        self._CurrentFrame = frame
        if 'position' in use: self.Position = key.PositionWorld
        if 'rotation' in use: self.Rotation = key.RotationWorld
        if 'scale' in use: self.Scale = key.ScaleWorld

        # may do it recursively
        if recursive:
            for child in self.Children:
                child.loadPose(frame, recursive=True, use=use)

        return self

    def writePose(self, frameId: int, recursive: bool = True) -> "Joint":
        """Sets joint properties as animation pose for the given frame id.
        - If there is already a keyframe at the frame id, it will be overwritten.
        - Inserts a new keyframe if there is none yet.
        - If the frame number is negative, it counts as the n-th frame from the end.
        - If recursive is True -> Child joints do also write their pose.

        Returns itself."""
        # add keyframe
        self.setKeyframe(frameId, pose=self, keep=['position', 'rotation', 'scale'])

        # recursion
        if recursive:
            for child in self.Children:
                child.writePose(frameId, recursive=True)

        return self

    def getKeyframeRange(self, includeChildren: bool = True) -> tuple[int, int]:
        """Returns the earliest and latest frame id of the animation.
        - If there are no keyframes, `(0, 0)` is returned.
        - If includeChildren is True -> The range considers the earliest and latest frames from its children too.
        The tuple layout is -> [FirstFrameId, LastFrameId]"""
        if len(self.Keyframes) == 0: return (0, 0)
        range = (self.Keyframes[0][0], self.Keyframes[-1][0])

        if includeChildren:
            for child in self.Children:
                childRange = child.getKeyframeRange(includeChildren=True)
                range = (min(range[0], childRange[0]), max(range[1], childRange[1]))

        return range

    def roll(self, degrees: float, recursive: bool = False) -> "Joint":
        """Rotates the joint along its local Y axis and updates the children so there is no spatial change.
        - RestPose and Keyframe data are not modified.

        Returns itself.
        """
        change = glm.angleAxis(glm.radians(degrees), (0, 1, 0))
        changeInverse = glm.inverse(change)

        self.Rotation = self.Rotation * change
        for child in self.Children:
            child.Position = changeInverse * child.Position
            child.Rotation = changeInverse * child.Rotation

            if recursive:
                child.roll(degrees, recursive=True)

        return self

    def attach(self, *nodes: "Joint", keep: list[str] = ['position', 'rotation', 'scale']) -> "Joint":
        return super().attach(*nodes, keep=keep)

    def detach(self, *nodes: "Joint", keep: list[str] = ['position', 'rotation', 'scale']) -> "Joint":
        return super().detach(*nodes, keep=keep)

    def clearParent(self, keep: list[str] = ['position', 'rotation', 'scale']) -> "Joint":
        return super().clearParent(keep=keep)

    def clearChildren(self, keep: list[str] = ['position', 'rotation', 'scale']) -> "Joint":
        return super().clearChildren(keep=keep)

    def applyPosition(self, position: glm.vec3 = None, recursive: bool = False) -> "Joint":
        return super().applyPosition(position, recursive)

    def applyRestposePosition(self, position: glm.vec3 = None, recursive: bool = False) -> "Joint":
        """"Resets the position of the Restpose to (0,0,0) or adds the given position.
        - This will load the restpose and overwrites the current pose of the transform!
        - You may want to call `loadPose` after this one or store the current properties.
        - Updates its keyframes position ONLY. Rotation and scale will be unchanged.
        - Updates its children Restposes positions to be spatially unchanged.
        - This does not update the childrens keyframes.

        Returns itself.
        """
        change, changeInverse = self.RestPose._applyPositionGetChanges(position)
        self.RestPose.applyPosition(position, recursive=False)

        for child in self.Children:
            child.RestPose._applyPositionChangeInverse(changeInverse)

            if recursive:
                child.applyRestposePosition(position=position, recursive=True)

        return self

    def applyRotation(self, rotation: glm.quat = None, recursive: bool = False, bake: bool = False) -> "Joint":
        return super().applyRotation(rotation, recursive, bake)

    def applyRestposeRotation(self, rotation: glm.quat = None, recursive: bool = False, bake: bool = False, bakeKeyframes: bool = False) -> "Joint":
        """"Resets the rotation of the Restpose to (1,0,0,0) or adds the given rotation.
        - This does not update the childrens keyframes.
        - If bake is True -> The childrens Restposes will change in position ONLY.
        - If bakeKeyframes is True -> Updates its keyframes position ONLY.

        Returns itself.
        """
        change, changeInverse = self.RestPose._applyRotationGetChanges(rotation)
        self.RestPose.applyRotation(rotation, recursive=False, bake=bakeKeyframes)

        for child in self.Children:
            child.RestPose._applyRotationChangeInverse(changeInverse, bake=bake)

            if recursive:
                child.applyRestposeRotation(rotation=rotation, recursive=True, bake=bake, bakeKeyframes=bakeKeyframes)

        return self

    def applyScale(self, scale: glm.vec3 = None, recursive: bool = False, bake: bool = False) -> "Joint":
        return super().applyScale(scale, recursive, bake)

    def applyRestposeScale(self, scale: glm.vec3 = None, recursive: bool = False, bake: bool = False, bakeKeyframes: bool = False) -> "Joint":
        """"Resets the scale of the Restpose to (1,1,1) or adds the given scale.
        - This does not update the childrens keyframes.
        - If bake is True -> The childrens Restposes will change in position ONLY.
        - If bakeKeyframes is True -> Updates its keyframes position ONLY.

        Returns itself.
        """
        change, changeInverse = self.RestPose._applyScaleGetChanges(scale)
        self.RestPose.applyScale(scale, recursive=False, bake=bakeKeyframes)

        for child in self.Children:
            child.RestPose._applyScaleChangeInverse(changeInverse, bake=bake)

            if recursive:
                child.applyRestposeScale(scale=scale, recursive=True, bake=bake, bakeKeyframes=bakeKeyframes)

        return self

    def setEuler(self, degrees: glm.vec3, order: str = 'ZXY', extrinsic: bool = True) -> "Joint":
        return super().setEuler(degrees, order, extrinsic)

    def filter(self, pattern: str, isEqual: bool = False, caseSensitive: bool = False) -> list["Joint"]:
        return super().filter(pattern, isEqual, caseSensitive)

    def filterRegex(self, pattern: str) -> list["Joint"]:
        return super().filterRegex(pattern)

    def layout(self, index: int = 0, depth: int = 0) -> list[tuple["Joint", int, int]]:
        return super().layout(index, depth)



import glm
import math
import random

deltaPosition = 1e-05
deltaRotation = 1e-05
deltaScale = 1e-05
randomSamples = 10000


def randomPosition():
    return glm.vec3(random.random(), random.random(), random.random()) * 2 - 1

def randomDirection():
    result = glm.vec3(random.random(), random.random(), random.random())
    theta = result.x * math.pi * 2.0
    phi = math.acos((2.0 * result.y) - 1.0)
    r = pow(result.z, 1/3)

    result.x = r * math.sin(phi) * math.cos(theta)
    result.y = r * math.sin(phi) * math.sin(theta)
    result.z = r * math.cos(phi)

    return glm.normalize(result)

def randomRotation():
    return glm.normalize(glm.quatLookAt(randomDirection(), (0,1,0)))

def randomScale():
    return (glm.vec3(random.random(), random.random(), random.random()) + 0.01) * \
         glm.vec3(-1 if random.random()<0.5 else 1, -1 if random.random()<0.5 else 1, -1 if random.random()<0.5 else 1)


def deviationPosition(a:glm.vec3, b:glm.vec3):
    return glm.l1Norm(a - b)

def deviationQuaternion(a:glm.quat, b:glm.quat):
    return sum([abs(d) for d in (a - b).to_list()])

def deviationScale(a:glm.vec3, b:glm.vec3):
    return glm.l1Norm(a - b)



from .BvhJoint import BvhJoint


class BvhContainer:
    """Container for the information of the bvh file.

    Root is the recursive definition of the skeleton.

    Frame time the frame time.

    Frams are the count of keyframes of the motion."""
    Root: BvhJoint
    FrameCount: int
    FrameTime: float

    def __init__(self, root: BvhJoint = None, frameCount: int = None, frameTime: float = None):
        self.Root: BvhJoint = root
        self.FrameCount: int = frameCount
        self.FrameTime: float = frameTime



from .lib.bvh import BvhContainer, BvhJoint
from .lib.hierarchy import Joint
from .lib.Parser import convertBvhToHierarchy, convertHierarchyToBvh, readAsHierarchy, readAsBvh, writeBvh, writeHierarchy
from SpatialTransform import Euler, Pose, Transform



import unittest
import bvhio

class Parser(unittest.TestCase):
    def test_readAsBVH(self):
        data = bvhio.readAsBvh('bvhio/tests/example.bvh')
        self.assertEqual(bvhio.BvhContainer, type(data))
        self.assertEqual(bvhio.BvhJoint, type(data.Root))
        self.assertEqual(data.FrameCount, 2)
        self.assertEqual(data.FrameTime, 0.033333)

    def test_readAsHierarchy(self):
        data = bvhio.readAsHierarchy('bvhio/tests/example.bvh')
        self.assertEqual(bvhio.Joint, type(data))
